@@ -43,7 +43,7 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
   private Params params;
   private JTabbedPane tabPane;
   private JPanel buttonPane, encodingPanel, connPanel, globalPanel, secPanel;
-  private JCheckBox jpeg, interframe;
+  private JCheckBox jpeg, interframe, turboVNCServer;
   private JComboBox menuKey, scalingFactor, encMethodComboBox, span,
     desktopSize, grabKeyboard;
   private JSlider jpegQualityLevel, subsamplingLevel, compressionLevel;
@@ -54,19 +54,18 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
     secTLSPlain, secX509None, secX509Vnc, secX509Plain, secUnixLogin;
   private JPanel encNonePanel, encTLSPanel, encX509Panel;
   private JCheckBox sendLocalUsername, tunnel;
-  private JTextField username, sshUser, gateway;
-  private JLabel usernameLabel, sshUserLabel, gatewayLabel;
+  private JTextField username, jump;
+  private JLabel usernameLabel, jumpLabel, jumpLabel2;
   private JButton okButton, cancelButton;
   private JButton x509caButton, x509crlButton;
   private JLabel x509caLabel, x509crlLabel;
   private JTextField x509ca, x509crl;
-  private JButton defClearButton, resetButton;
+  private JButton listClearButton, defClearButton, resetButton;
   private JLabel encMethodLabel;
   private JLabel jpegQualityLabel, jpegQualityLabelLo, jpegQualityLabelHi;
   private JLabel subsamplingLabel, subsamplingLabelLo, subsamplingLabelHi;
-  private JLabel compressionLabel, compressionLabelLo, compressionLabelHi;
+  private JLabel compressionLabel;
   private String jpegQualityLabelString, subsamplingLabelString;
-  private String compressionLabelString;
   private Hashtable<Integer, String> subsamplingLabelTable;
   private String oldScalingFactor, oldDesktopSize;
   private boolean enableX509 = true;
@@ -193,8 +192,6 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
                           GridBagConstraints.LINE_START,
                           new Insets(2, 2, 2, 2));
 
-    compressionLabelString =
-      new String("Compression level (see docs): ");
     compressionLabel = new JLabel();
     compressionLevel =
       new JSlider(JSlider.HORIZONTAL, 0, 2, 1);
@@ -203,9 +200,7 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
     compressionLevel.setSnapToTicks(true);
     compressionLevel.setPaintTicks(true);
     compressionLevel.setPaintLabels(false);
-    compressionLabelLo = new JLabel("fast");
-    compressionLabelHi = new JLabel("best");
-    compressionLabel.setText(compressionLabelString +
+    compressionLabel.setText("Compression level (see docs): " +
                              compressionLevel.getValue());
 
     Dialog.addGBComponent(compressionLabel, dummyPanel,
@@ -213,18 +208,8 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
                           GridBagConstraints.NONE,
                           GridBagConstraints.LINE_START,
                           new Insets(2, 2, 2, 2));
-    Dialog.addGBComponent(compressionLabelLo, dummyPanel,
-                          0, 8, 1, 1, 2, 2, 0, 0,
-                          GridBagConstraints.NONE,
-                          GridBagConstraints.LINE_START,
-                          new Insets(2, 12, 2, 2));
     Dialog.addGBComponent(compressionLevel, dummyPanel,
-                          1, 8, 1, 1, 2, 2, 0, 0,
-                          GridBagConstraints.NONE,
-                          GridBagConstraints.LINE_START,
-                          new Insets(2, 2, 2, 2));
-    Dialog.addGBComponent(compressionLabelHi, dummyPanel,
-                          2, 8, 1, 1, 2, 2, 1, 1,
+                          0, 8, 3, 1, 2, 2, 0, 0,
                           GridBagConstraints.NONE,
                           GridBagConstraints.LINE_START,
                           new Insets(2, 2, 2, 2));
@@ -238,10 +223,19 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
                           GridBagConstraints.LINE_START,
                           new Insets(2, 0, 2, 2));
 
+    turboVNCServer = new JCheckBox("TurboVNC server");
+    turboVNCServer.addItemListener(this);
+
+    Dialog.addGBComponent(turboVNCServer, dummyPanel,
+                          0, 10, 3, 1, 0, 2, 0, 0,
+                          GridBagConstraints.NONE,
+                          GridBagConstraints.LINE_START,
+                          new Insets(2, 0, 2, 2));
+
     Dialog.addGBComponent(dummyPanel, encodingPanel,
                           0, 0, 1, GridBagConstraints.REMAINDER, 0, 0, 1, 1,
-                          GridBagConstraints.HORIZONTAL,
-                          GridBagConstraints.PAGE_START,
+                          GridBagConstraints.NONE,
+                          GridBagConstraints.FIRST_LINE_START,
                           new Insets(4, 4, 4, 4));
 
     // Connection tab
@@ -382,7 +376,6 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
 
     viewOnly = new JCheckBox("View only (ignore mouse & keyboard)");
     viewOnly.addItemListener(this);
-    viewOnly.setEnabled(params.viewOnlyControl.get());
 
     Dialog.addGBComponent(viewOnly, inputPanel,
                           0, 0, 2, 1, 2, 2, 1, 0,
@@ -506,18 +499,25 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
     // Global tab
     globalPanel = new JPanel(new GridBagLayout());
 
-    defClearButton = new JButton("Clear the list of saved connections");
+    listClearButton = new JButton("Clear the list of saved connections");
+    listClearButton.addActionListener(this);
+    defClearButton = new JButton("Clear all saved per-host options");
     defClearButton.addActionListener(this);
     resetButton = new JButton("Reset all options to defaults");
     resetButton.addActionListener(this);
 
-    Dialog.addGBComponent(defClearButton, globalPanel,
+    Dialog.addGBComponent(listClearButton, globalPanel,
                           0, 0, 2, 1, 2, 2, 1, 0,
                           GridBagConstraints.NONE,
                           GridBagConstraints.FIRST_LINE_START,
                           new Insets(8, 5, 0, 5));
+    Dialog.addGBComponent(defClearButton, globalPanel,
+                          0, 1, 2, 1, 2, 2, 1, 0,
+                          GridBagConstraints.NONE,
+                          GridBagConstraints.FIRST_LINE_START,
+                          new Insets(8, 5, 0, 5));
     Dialog.addGBComponent(resetButton, globalPanel,
-                          0, 1, 2, GridBagConstraints.REMAINDER, 2, 2, 1, 1,
+                          0, 2, 2, GridBagConstraints.REMAINDER, 2, 2, 1, 1,
                           GridBagConstraints.NONE,
                           GridBagConstraints.FIRST_LINE_START,
                           new Insets(8, 5, 0, 5));
@@ -667,39 +667,35 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
     sendLocalUsername = new JCheckBox("Send local username");
     sendLocalUsername.addItemListener(this);
 
-    JPanel gatewayPanel = new JPanel(new GridBagLayout());
-    gatewayPanel.setBorder(
-      BorderFactory.createTitledBorder("Gateway (SSH server or UltraVNC repeater)"));
-    gateway = new JTextField("", 1);
-    filterWhitespace(gateway);
-    gatewayLabel = new JLabel("Host:");
-    sshUser = new JTextField("", 1);
-    sshUserLabel = new JLabel("SSH user:");
-    tunnel = new JCheckBox("Use VNC server as gateway");
+    JPanel sshPanel = new JPanel(new GridBagLayout());
+    sshPanel.setBorder(BorderFactory.createTitledBorder("SSH tunneling"));
+    jumpLabel =
+      new JLabel("Tunnel VNC connection through specified SSH jump host");
+    jumpLabel2 =
+      new JLabel("([user@]host[:ssh_port]) and VNC host's SSH server");
+    jump = new JTextField("", 1);
+    filterWhitespace(jump);
+    tunnel =
+      new JCheckBox("Tunnel VNC connection through VNC host's SSH server");
     tunnel.addItemListener(this);
 
-    Dialog.addGBComponent(sshUserLabel, gatewayPanel,
-                          0, 0, 1, 1, 2, 2, 0, 0,
-                          GridBagConstraints.NONE,
-                          GridBagConstraints.FIRST_LINE_START,
-                          new Insets(8, 8, 0, 2));
-    Dialog.addGBComponent(sshUser, gatewayPanel,
-                          1, 0, 1, 1, 2, 2, 0.7, 0,
+    Dialog.addGBComponent(jumpLabel, sshPanel,
+                          0, 0, 1, 1, 0, 0, 1, 1,
                           GridBagConstraints.HORIZONTAL,
-                          GridBagConstraints.FIRST_LINE_START,
-                          new Insets(4, 2, 0, 2));
-    Dialog.addGBComponent(gatewayLabel, gatewayPanel,
-                          2, 0, 1, 1, 2, 2, 0, 0,
-                          GridBagConstraints.NONE,
-                          GridBagConstraints.FIRST_LINE_START,
-                          new Insets(8, 2, 0, 2));
-    Dialog.addGBComponent(gateway, gatewayPanel,
-                          3, 0, 1, 1, 2, 2, 1, 0,
+                          GridBagConstraints.CENTER,
+                          new Insets(0, 5, 0, 5));
+    Dialog.addGBComponent(jumpLabel2, sshPanel,
+                          0, 1, 1, 1, 0, 0, 1, 1,
+                          GridBagConstraints.HORIZONTAL,
+                          GridBagConstraints.CENTER,
+                          new Insets(0, 5, 0, 5));
+    Dialog.addGBComponent(jump, sshPanel,
+                          0, 2, 1, 1, 2, 2, 1, 0,
                           GridBagConstraints.HORIZONTAL,
                           GridBagConstraints.FIRST_LINE_START,
                           new Insets(4, 2, 0, 5));
-    Dialog.addGBComponent(tunnel, gatewayPanel,
-                          0, 1, 4, 1, 2, 2, 1, 1,
+    Dialog.addGBComponent(tunnel, sshPanel,
+                          0, 3, 4, 1, 2, 2, 1, 1,
                           GridBagConstraints.HORIZONTAL,
                           GridBagConstraints.FIRST_LINE_START,
                           new Insets(4, 5, 0, 5));
@@ -729,7 +725,7 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
                           GridBagConstraints.HORIZONTAL,
                           GridBagConstraints.FIRST_LINE_START,
                           new Insets(2, 10, 2, 5));
-    Dialog.addGBComponent(gatewayPanel, secPanel,
+    Dialog.addGBComponent(sshPanel, secPanel,
                           0, 3, 3, 1, 2, 2, 1, 1,
                           GridBagConstraints.HORIZONTAL,
                           GridBagConstraints.FIRST_LINE_START,
@@ -843,11 +839,14 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
       endDialog();
     } else if (s instanceof JButton && (JButton)s == cancelButton) {
       endDialog();
+    } else if (s instanceof JButton && (JButton)s == listClearButton) {
+      UserPreferences.clear(true);
     } else if (s instanceof JButton && (JButton)s == defClearButton) {
-      UserPreferences.clear();
+      UserPreferences.clear(false);
     } else if (s instanceof JButton && (JButton)s == resetButton) {
       Params oldParams = params;
       params = new Params();
+      params.loadDefaults();
       setOptions(desktopSize.isEnabled(), !shared.isEnabled(), false, false);
       params = oldParams;
     } else if (s instanceof JButton && (JButton)s == x509caButton) {
@@ -911,10 +910,6 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
         subsamplingLabel.setEnabled(true);
         subsamplingLabelLo.setEnabled(true);
         subsamplingLabelHi.setEnabled(true);
-        if (compressionLevel.getMaximum() < 9) {
-          compressionLevel.setMinimum(1);
-          compressionLevel.setMaximum(2);
-        }
       } else {
         jpegQualityLevel.setEnabled(false);
         jpegQualityLabel.setEnabled(false);
@@ -924,17 +919,18 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
         subsamplingLabel.setEnabled(false);
         subsamplingLabelLo.setEnabled(false);
         subsamplingLabelHi.setEnabled(false);
-        if (compressionLevel.getMaximum() < 9) {
-          compressionLevel.setMinimum(0);
-          compressionLevel.setMaximum(1);
-        }
       }
+      if (turboVNCServer.isSelected())
+        setCompatibleGUI();
       setEncMethodComboBox();
     }
     if (s instanceof JCheckBox && (JCheckBox)s == interframe) {
+      if (turboVNCServer.isSelected())
+        setCompatibleGUI();
       setEncMethodComboBox();
-      compressionLabel.setText(compressionLabelString + getCompressionLevel());
     }
+    if (s instanceof JCheckBox && (JCheckBox)s == turboVNCServer)
+      setCompressionLevel(getCompressionLevel());
     if (s instanceof JCheckBox && (JCheckBox)s == secNone ||
         s instanceof JCheckBox && (JCheckBox)s == secVnc ||
         s instanceof JCheckBox && (JCheckBox)s == secPlain ||
@@ -991,8 +987,9 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
       }
     }
     if (s instanceof JCheckBox && (JCheckBox)s == tunnel) {
-      gateway.setEnabled(!tunnel.isSelected());
-      gatewayLabel.setEnabled(!tunnel.isSelected());
+      jump.setEnabled(!tunnel.isSelected());
+      jumpLabel.setEnabled(!tunnel.isSelected());
+      jumpLabel2.setEnabled(!tunnel.isSelected());
     }
   }
 
@@ -1055,7 +1052,7 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
         jpegQualityLevel.getValue());
       setEncMethodComboBox();
     } else if (s instanceof JSlider && (JSlider)s == compressionLevel) {
-      compressionLabel.setText(compressionLabelString + getCompressionLevel());
+      setCompatibleGUI();
       setEncMethodComboBox();
     }
   }
@@ -1082,18 +1079,47 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
     return false;
   }
 
-  public void setCompressionLevel(int level) {
-    boolean selectICE = false;
-    if (!isTurboCompressionLevel(level)) {
+  void setCompatibleGUI() {
+    String compressionLabelString;
+
+    if (!turboVNCServer.isSelected()) {
       compressionLevel.setMinimum(0);
       compressionLevel.setMaximum(9);
       compressionLabelString = new String("Compression level: ");
-      interframe.setEnabled(false);
+    } else {
+      if (jpeg.isSelected()) {
+        compressionLevel.setMinimum(1);
+        compressionLevel.setMaximum(2);
+      } else {
+        compressionLevel.setMinimum(0);
+        compressionLevel.setMaximum(1);
+      }
+      compressionLabelString = new String("Compression level (see docs): ");
     }
-    if (compressionLevel.getMaximum() < 9 &&
-        interframe.isEnabled()) {
-      if (level >= 5 && level <= 8) {
-        level -= 5;
+    String clString = "";
+    int cl = getCompressionLevel();
+    if (turboVNCServer.isSelected()) {
+      if (cl == 0 || cl == 5) clString = " [no zlib]";
+      else if (cl == 1 || cl == 6) clString = " [minimal]";
+      else if (cl == 2 || cl == 7) clString = " [medium]";
+    } else {
+      if (cl == 0) clString = " [none]";
+      else if (cl == 1) clString = " [minimal]";
+      else if (cl == 2) clString = " [medium]";
+      else if (cl == 6) clString = " [maximum useful]";
+    }
+    if (cl == 9) clString = " [extreme]";
+    compressionLabel.setText(compressionLabelString + cl + clString);
+    interframe.setEnabled(turboVNCServer.isSelected());
+  }
+
+  void setCompressionLevel(int level) {
+    boolean selectICE = false;
+    setCompatibleGUI();
+    if (turboVNCServer.isSelected() && interframe.isEnabled()) {
+      if (level >= 5 && level <= 9) {
+        if (level <= 8)
+          level -= 5;
         selectICE = true;
       }
     }
@@ -1134,13 +1160,7 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
       encMethodComboBox.setSelectedItem(RFB.encodingName(encoding));
       encMethodLabel.setText("Encoding type:");
       encMethodLabel.setEnabled(false);
-    }
-    if (encoding != RFB.ENCODING_TIGHT || params.compatibleGUI.get()) {
-      compressionLevel.setMinimum(0);
-      compressionLevel.setMaximum(9);
-      compressionLabelString = new String("Compression level: ");
-      compressionLabel.setText(compressionLabelString + getCompressionLevel());
-      interframe.setEnabled(false);
+      turboVNCServer.setEnabled(false);
     }
   }
 
@@ -1206,6 +1226,8 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
     jpeg.setSelected(params.jpeg.get());
     subsamplingLevel.setValue(params.subsampling.getOrdinal());
     jpegQualityLevel.setValue(params.quality.get());
+    turboVNCServer.setSelected(!params.compatibleGUI.get() &&
+      isTurboCompressionLevel(params.compressLevel.get()));
     setCompressionLevel(params.compressLevel.get());
 
     // Connection: Display
@@ -1263,9 +1285,8 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
     x509ca.setText(params.x509ca.get());
     x509crl.setText(params.x509crl.get());
 
-    // Security: Gateway
-    sshUser.setText(params.sshUser.get());
-    gateway.setText(params.via.get());
+    // Security: SSH tunneling
+    jump.setText(params.jump.get());
     tunnel.setSelected(params.tunnel.get());
 
     desktopSize.setEnabled(enableDesktopSize);
@@ -1297,10 +1318,9 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
       x509crlLabel.setEnabled(false);
     }
     if (disableSSH) {
-      sshUser.setEnabled(false);
-      sshUserLabel.setEnabled(false);
-      gateway.setEnabled(false);
-      gatewayLabel.setEnabled(false);
+      jumpLabel.setEnabled(false);
+      jumpLabel2.setEnabled(false);
+      jump.setEnabled(false);
       tunnel.setEnabled(false);
     }
   }
@@ -1311,6 +1331,7 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
     params.subsampling.set(getSubsamplingLevel());
     params.quality.set(jpegQualityLevel.getValue());
     params.compressLevel.set(getCompressionLevel());
+    params.compatibleGUI.set(!turboVNCServer.isSelected());
 
     // Connection: Display
     params.scale.set(scalingFactor.getSelectedItem().toString());
@@ -1375,9 +1396,8 @@ class OptionsDialog extends Dialog implements ActionListener, ChangeListener,
     params.x509ca.set(x509ca.getText().isEmpty() ? null : x509ca.getText());
     params.x509crl.set(x509crl.getText().isEmpty() ? null : x509crl.getText());
 
-    // Security: Gateway
-    params.sshUser.set(sshUser.getText().isEmpty() ? null : sshUser.getText());
-    params.via.set(gateway.getText().isEmpty() ? null : gateway.getText());
+    // Security: SSH tunneling
+    params.jump.set(jump.getText().isEmpty() ? null : jump.getText());
     params.tunnel.set(tunnel.isSelected());
 
     params.reconcile();
