@@ -34,67 +34,98 @@ import com.turbovnc.rfb.*;
 import com.turbovnc.rdr.*;
 import com.turbovnc.network.*;
 
-import com.jcraft.jsch.agentproxy.*;
-import com.jcraft.jsch.agentproxy.connector.*;
-import com.jcraft.jsch.agentproxy.usocket.*;
-import com.jcraft.jsch.Identity;
 import com.jcraft.jsch.*;
 
 public class Tunnel {
 
   public static void createTunnel(Params params) throws Exception {
-    int localPort, remotePort;
-    String gatewayHost, remoteHost;
+    int vncPort;
 
-    boolean tunnel = params.tunnel.get() ||
-                     (params.sessMgrActive && params.sessMgrAuto.get());
-
-    if (tunnel) {
-      gatewayHost = Hostname.getHost(params.server.get());
-      remoteHost = "localhost";
-    } else {
-      gatewayHost = params.via.get();
-      remoteHost = Hostname.getHost(params.server.get());
-    }
     if (params.server.get() != null &&
         Hostname.getColonPos(params.server.get()) < 0 && params.port.get() > 0)
-      remotePort = params.port.get();
+      vncPort = params.port.get();
     else
-      remotePort = Hostname.getPort(params.server.get());
+      vncPort = Hostname.getPort(params.server.get());
 
-    String pattern = null;
-    if (tunnel) {
-      pattern = System.getProperty("turbovnc.tunnel");
-      if (pattern == null)
-        pattern = System.getenv("VNC_TUNNEL_CMD");
-    } else {
-      pattern = System.getProperty("turbovnc.via");
-      if (pattern == null)
-        pattern = System.getenv("VNC_VIA_CMD");
+    String pattern = params.extSSHTemplate.get();
+    if (pattern == null) {
+      if (params.via.get() != null) {
+        pattern = System.getProperty("turbovnc.via");
+        if (pattern == null)
+          pattern = System.getenv("VNC_VIA_CMD");
+      } else {
+        pattern = System.getProperty("turbovnc.tunnel");
+        if (pattern == null)
+          pattern = System.getenv("VNC_TUNNEL_CMD");
+      }
     }
 
     if (params.udsPath != null &&
         (pattern == null || pattern.indexOf("%L") < 0)) {
       // Connect to Unix domain socket using stdio-based forwarding
-      params.stdioSocket = createTunnelExtUDS(gatewayHost, remoteHost, pattern,
-                                              params, tunnel);
+      params.stdioSocket = createTunnelExtUDS(pattern, params);
     } else {
-      localPort = TcpSocket.findFreeTcpPort();
+      int localPort = TcpSocket.findFreeTcpPort();
       if (localPort == 0)
         throw new ErrorException("Could not obtain free TCP port");
 
       if (params.extSSH.get() || (pattern != null && pattern.length() > 0))
-        createTunnelExt(gatewayHost, remoteHost, remotePort, localPort,
-                        pattern, params, tunnel);
+        createTunnelExt(vncPort, localPort, pattern, params);
       else {
-        vlog.debug("Opening SSH tunnel through gateway " + gatewayHost);
-        if (params.sshSession == null)
-          createTunnelJSch(gatewayHost, params);
-        remoteHost = remoteHost.replaceAll("[\\[\\]]", "");
-        vlog.debug("Forwarding local port " + localPort + " to " + remoteHost +
-                   "::" + remotePort + " (relative to gateway)");
-        params.sshSession.setPortForwardingL(localPort, remoteHost,
-                                             remotePort);
+        String vncHost, gatewayStr;
+        int sshPort = params.sshPort.isDefault() ? -1 : params.sshPort.get();
+
+        if (params.jump.get() != null && !params.tunnel.get()) {
+          gatewayStr = Hostname.getHost(params.server.get());
+          if (params.sshSession == null) {
+            vncHost = gatewayStr;
+            vlog.debug("Opening SSH connection to host " + vncHost);
+            params.sshSession =
+              createTunnelJSch(Hostname.getSSHUser(params.server.get()),
+                               vncHost, sshPort,
+                               Hostname.getSSHUser(params.jump.get()),
+                               Hostname.getSSHHost(params.jump.get()),
+                               Hostname.getSSHPort(params.jump.get()), params);
+          }
+          vncHost = "localhost";
+        } else {
+          String gatewaySSHUser, gatewayHost;
+
+          boolean tunnel = params.tunnel.get() ||
+                           (params.sessMgrActive && params.sessMgrAuto.get());
+
+          if (tunnel) {
+            gatewaySSHUser = Hostname.getSSHUser(params.server.get());
+            gatewayHost = Hostname.getHost(params.server.get());
+            vncHost = "localhost";
+          } else {
+            gatewaySSHUser = Hostname.getSSHUser(params.via.get());
+            gatewayHost = Hostname.getHost(params.via.get());
+            vncHost = Hostname.getHost(params.server.get());
+            // If the Session Manager is active and SessMgrAuto=0, then Via
+            // uses multi-level SSH tunneling for the Session Manager's SSH
+            // connection, but it uses direct port forwarding for the RFB/SSH
+            // connection.  Thus, close the Session Manager's multi-level SSH
+            // connection so a new single-level SSH connection can be created
+            // below.
+            if (params.sessMgrActive && params.sshSession != null) {
+              params.sshSession.disconnect();
+              params.sshSession = null;
+            }
+          }
+
+          if (params.sshSession == null) {
+            vlog.debug("Opening SSH tunnel through gateway " + gatewayHost);
+            params.sshSession = createTunnelJSch(gatewaySSHUser, gatewayHost,
+                                                 sshPort, null, null, -1,
+                                                 params);
+          }
+          vncHost = vncHost.replaceAll("[\\[\\]]", "");
+          gatewayStr = gatewayHost;
+        }
+        vlog.debug("Forwarding local port " + localPort + " to " + vncHost +
+                   "::" + vncPort + " (relative to " + gatewayStr + ")");
+        params.sshSession.setPortForwardingL(localPort, vncHost, vncPort);
       }
       params.server.set("localhost::" + localPort);
     }
@@ -103,8 +134,56 @@ public class Tunnel {
 
   /* Create a tunnel using the built-in JSch SSH client */
 
-  protected static void createTunnelJSch(String host, Params params)
-                                         throws Exception {
+  private static class JumpProxy implements Proxy {
+    Session jumpSSHSession;
+    ChannelDirectTCPIP channel;
+    PipedOutputStream osToRemote;
+    PipedInputStream isFromRemote;
+
+    JumpProxy(Session jumpSSHSession_) {
+      jumpSSHSession = jumpSSHSession_;
+    }
+
+    public void connect(SocketFactory socket_factory, String host, int port,
+                        int timeout) throws Exception {
+      channel = (ChannelDirectTCPIP)jumpSSHSession.openChannel("direct-tcpip");
+      channel.setHost(host);
+      channel.setPort(port);
+      PipedInputStream isToRemote = new PipedInputStream();
+      osToRemote = new PipedOutputStream(isToRemote);
+      isFromRemote = new PipedInputStream();
+      PipedOutputStream osFromRemote = new PipedOutputStream(isFromRemote);
+      channel.setInputStream(isToRemote);
+      channel.setOutputStream(osFromRemote);
+      channel.connect();
+    }
+
+    public InputStream getInputStream() {
+      return isFromRemote;
+    }
+
+    public OutputStream getOutputStream() {
+      return osToRemote;
+    }
+
+    public java.net.Socket getSocket() {
+      return null;
+    }
+
+    public void close() {
+      if (channel != null)
+        channel.disconnect();
+      channel = null;
+      isFromRemote = null;
+      osToRemote = null;
+      jumpSSHSession.disconnect();
+    }
+  }
+
+  protected static Session createTunnelJSch(String user, String host, int port,
+                                            String jumpUser, String jumpHost,
+                                            int jumpPort, Params params)
+                                            throws Exception {
     JSch jsch = new JSch();
     JSch.setLogger(LOGGER);
     String homeDir = new String("");
@@ -122,14 +201,14 @@ public class Tunnel {
     jsch.setKnownHosts(knownHosts.getAbsolutePath());
 
     if (Helper.isAvailable()) {
-      Connector connector = null;
+      AgentConnector connector = null;
       try {
         if (Utils.isWindows())
           connector = new PageantConnector();
         else
-          connector = new SSHAgentConnector(new JNIUSocketFactory());
+          connector = new SSHAgentConnector(new UnixDomainSocketFactory());
         if (connector != null) {
-          IdentityRepository repo = new RemoteIdentityRepository(connector);
+          IdentityRepository repo = new AgentIdentityRepository(connector);
           vlog.sshdebug("SSH private keys offered by agent:");
           Iterator<Identity> iter = repo.getIdentities().iterator();
           while (iter.hasNext()) {
@@ -174,9 +253,6 @@ public class Tunnel {
     }
 
     // username and passphrase will be given via UserInfo interface.
-    int port = params.sshPort.get();
-    String user = params.sshUser.get();
-
     File sshConfigFile = new File(params.sshConfig.get());
     if (sshConfigFile.exists() && sshConfigFile.canRead()) {
       ConfigRepository repo =
@@ -187,7 +263,7 @@ public class Tunnel {
       if (repoUser != null && user == null)
         user = repoUser;
       int repoPort = repo.getConfig(host).getPort();
-      if (repoPort != -1 && params.sshPort.isDefault())
+      if (repoPort != -1 && port == -1)
         port = repoPort;
       String[] identityFiles = repo.getConfig(host).getValues("IdentityFile");
       if (identityFiles != null) {
@@ -196,6 +272,14 @@ public class Tunnel {
             useDefaultPrivateKeyFiles = false;
         }
       }
+
+      String repoJumpHost = repo.getConfig(host).getValue("ProxyJump");
+      if (repoJumpHost != null && jumpHost == null) {
+        jumpHost = repoJumpHost.replaceAll("\\s", "");
+        jumpUser = Hostname.getSSHUser(jumpHost);
+        jumpPort = Hostname.getSSHPort(jumpHost);
+        jumpHost = Hostname.getSSHHost(jumpHost);
+      }
     } else {
       if (params.sshConfig.isDefault()) {
         vlog.debug("Could not parse SSH config file " +
@@ -213,8 +297,9 @@ public class Tunnel {
 
     if (useDefaultPrivateKeyFiles) {
       privateKeys.add(new File(homeDir + "/.ssh/id_rsa"));
-      privateKeys.add(new File(homeDir + "/.ssh/id_dsa"));
       privateKeys.add(new File(homeDir + "/.ssh/id_ecdsa"));
+      privateKeys.add(new File(homeDir + "/.ssh/id_ed25519"));
+      privateKeys.add(new File(homeDir + "/.ssh/id_dsa"));
     }
 
     for (Iterator<File> i = privateKeys.iterator(); i.hasNext();) {
@@ -234,11 +319,23 @@ public class Tunnel {
       }
     }
 
-    params.sshSession = jsch.getSession(user, host, port);
+    Session sshSession = null, jumpSSHSession = null;
+
+    if (jumpHost != null) {
+      vlog.debug("Opening SSH connection to jump host " + jumpHost);
+      jumpSSHSession = createTunnelJSch(jumpUser, jumpHost, jumpPort, null,
+                                        null, -1, params);
+    }
+    if (port == -1) port = params.sshPort.getDefault();
+    sshSession = jsch.getSession(user, host, port);
+    if (jumpSSHSession != null) {
+      JumpProxy proxy = new JumpProxy(jumpSSHSession);
+      sshSession.setProxy(proxy);
+    }
 
     vlog.sshdebug("Attempting to use the following SSH private keys:");
     Iterator<Identity> iter =
-      params.sshSession.getIdentityRepository().getIdentities().iterator();
+      sshSession.getIdentityRepository().getIdentities().iterator();
     while (iter.hasNext()) {
       Identity id = iter.next();
       vlog.sshdebug("  " + id.getName());
@@ -247,39 +344,44 @@ public class Tunnel {
     }
 
     // OpenSSHConfig doesn't recognize StrictHostKeyChecking
-    if (params.sshSession.getConfig("StrictHostKeyChecking") == null)
-      params.sshSession.setConfig("StrictHostKeyChecking", "ask");
-    params.sshSession.setConfig("MaxAuthTries", "3");
+    if (sshSession.getConfig("StrictHostKeyChecking") == null)
+      sshSession.setConfig("StrictHostKeyChecking", "ask");
+    sshSession.setConfig("MaxAuthTries", "3");
     String auth = System.getProperty("turbovnc.sshauth");
     if (auth != null)
-      params.sshSession.setConfig("PreferredAuthentications", auth);
+      sshSession.setConfig("PreferredAuthentications", auth);
     PasswdDialog dlg = new PasswdDialog(new String("SSH Authentication"),
                                         true, user, false, true, -1);
     if (!Utils.getBooleanProperty("turbovnc.sshkeytest", false))
-      params.sshSession.setUserInfo(dlg);
-    params.sshSession.connect();
+      sshSession.setUserInfo(dlg);
+    sshSession.connect();
+
+    return sshSession;
   }
 
-  /* Create a tunnel using an external SSH client.  This supports the same
-     VNC_TUNNEL_CMD and VNC_VIA_CMD environment variables as the native viewers
-     do. */
+  /* Create a tunnel using an external SSH client. */
 
-  private static final String DEFAULT_TUNNEL_CMD =
+  public static final String DEFAULT_TUNNEL_CMD =
     "%S -f -L %L:localhost:%R %H sleep 20";
-  private static final String DEFAULT_VIA_CMD =
+  public static final String DEFAULT_JUMP_CMD =
+    "%S -J %G -f -L %L:localhost:%R %H sleep 20";
+  public static final String DEFAULT_VIA_CMD =
     "%S -f -L %L:%H:%R %G sleep 20";
 
-  private static void createTunnelExt(String gatewayHost, String remoteHost,
-                                      int remotePort, int localPort,
-                                      String pattern, Params params,
-                                      boolean tunnel)
+  private static void createTunnelExt(int vncPort, int localPort,
+                                      String pattern, Params params)
                                       throws Exception {
-    if (pattern == null || pattern.length() < 1)
-      pattern = (tunnel ? DEFAULT_TUNNEL_CMD : DEFAULT_VIA_CMD);
+    boolean tunnel = params.tunnel.get() ||
+                     (params.sessMgrActive && params.sessMgrAuto.get());
 
-    String command = fillCmdPattern(pattern, gatewayHost, remoteHost,
-                                    remotePort, params.udsPath, localPort,
-                                    params, tunnel);
+    if (pattern == null || pattern.length() < 1)
+      pattern =
+        (tunnel ? DEFAULT_TUNNEL_CMD :
+                  (params.jump.get() != null ? DEFAULT_JUMP_CMD :
+                                               DEFAULT_VIA_CMD));
+
+    String command = fillCmdPattern(pattern, vncPort, params.udsPath,
+                                    localPort, params);
 
     vlog.debug("SSH command line: " + command);
     List<String> args = ArgumentTokenizer.tokenize(command);
@@ -292,25 +394,23 @@ public class Tunnel {
       throw new ErrorException("External SSH error");
   }
 
-  private static final String DEFAULT_TUNNEL_CMD_UDS =
+  public static final String DEFAULT_TUNNEL_CMD_UDS =
     "%S -- %H exec socat stdio unix-connect:%R";
-  private static final String DEFAULT_VIA_CMD_UDS =
+  public static final String DEFAULT_JUMP_CMD_UDS =
     "%S -J %G -- %H exec socat stdio unix-connect:%R";
 
-  private static Socket createTunnelExtUDS(String gatewayHost,
-                                           String remoteHost, String pattern,
-                                           Params params, boolean tunnel)
+  private static Socket createTunnelExtUDS(String pattern, Params params)
                                            throws Exception {
     if (pattern == null || pattern.length() < 1)
-      pattern = (tunnel ? DEFAULT_TUNNEL_CMD_UDS : DEFAULT_VIA_CMD_UDS);
+      pattern = (params.tunnel.get() ? DEFAULT_TUNNEL_CMD_UDS :
+                                       DEFAULT_JUMP_CMD_UDS);
 
     // Escape the Unix domain socket path twice, since it will be interpreted
     // once by ArgumentTokenizer.tokenize() and again by the remote shell.
     String udsPath = escapeUDSPath(params.udsPath, true);
     udsPath = escapeUDSPath(udsPath, false);
 
-    String command = fillCmdPattern(pattern, gatewayHost, remoteHost, -1,
-                                    udsPath, -1, params, tunnel);
+    String command = fillCmdPattern(pattern, -1, udsPath, -1, params);
 
     vlog.debug("SSH command line (stdio): " + command);
     List<String> args = ArgumentTokenizer.tokenize(command);
@@ -340,17 +440,13 @@ public class Tunnel {
     }
   }
 
-  private static String fillCmdPattern(String pattern, String gatewayHost,
-                                       String remoteHost, int remotePort,
+  private static String fillCmdPattern(String pattern, int vncPort,
                                        String udsPath, int localPort,
-                                       Params params, boolean tunnel) {
+                                       Params params) {
     int i, j;
     boolean hFound = false, gFound = false, rFound = false, lFound = false;
     String command = "";
 
-    if (params.sshUser.get() != null)
-      gatewayHost = params.sshUser.get() + "@" + gatewayHost;
-
     for (i = 0; i < pattern.length(); i++) {
       if (pattern.charAt(i) == '%') {
         switch (pattern.charAt(++i)) {
@@ -358,18 +454,24 @@ public class Tunnel {
             command += params.extSSHCommand.get();
             continue;
           case 'H':
-            command += (tunnel ? gatewayHost : remoteHost);
+            String vncSSHUser = Hostname.getSSHUser(params.server.get());
+            String vncHost = Hostname.getHost(params.server.get());
+            if (vncSSHUser != null)
+              command += vncSSHUser + "@" + vncHost;
+            else
+              command += vncHost;
             hFound = true;
             continue;
           case 'G':
-            command += gatewayHost;
+            command += (params.jump.get() != null ? params.jump.get() :
+                                                    params.via.get());
             gFound = true;
             continue;
           case 'R':
             if (udsPath != null)
               command += udsPath;
             else
-              command += remotePort;
+              command += vncPort;
             rFound = true;
             continue;
           case 'L':
@@ -384,7 +486,7 @@ public class Tunnel {
     if (!hFound || !rFound || (!lFound && udsPath == null))
       throw new ErrorException("%H, %R or %L absent in tunneling command template.");
 
-    if (!tunnel && !gFound)
+    if ((params.jump.get() != null || params.via.get() != null) && !gFound)
       throw new ErrorException("%G pattern absent in tunneling command template.");
 
     return command;
@@ -526,12 +628,12 @@ public class Tunnel {
           vlogSSH.error(message);
           return;
         case Logger.FATAL:
-          throw new ErrorException("JSch: " + message);
+          throw new ErrorException("SSH: " + message);
       }
     }
   };
 
   protected Tunnel() {}
   static LogWriter vlog = new LogWriter("Tunnel");
-  static LogWriter vlogSSH = new LogWriter("JSch");
+  static LogWriter vlogSSH = new LogWriter("SSH");
 }
